// File: MessagePass.java
// Pass a message among threads; variant of kids' operator game
// Need 3 classes here, a thread class, a main class to
//      instantiate some instances of the thread class
//      and a monitor class for synchronizing the threads
//
// Note that there are no throws clauses because every possible exception is caught

import java.io.*;

class Participant extends Thread
{	private int number;
 	private String Message;
 	Monitor Control;

 // Constructor: Receive thread id number and reference to monitor object
 public Participant(String Name,Monitor Controller)
 {	super(Name);
  	// Announce thread number and # milliseconds it will sleep
  	number=Integer.valueOf(getName()).intValue();
  	Control=Controller;
 }

 // run() executes when the thread is started
 public void run()
 {	try {
      		Control.waitForTurn(number);
      		Message=Control.getMessage();   // Get Message from Monitor
  	}
  	catch (Exception e) 
	{}

  	// Announce (this can be removed if only the final message is desired)
  	System.out.println("Thread "+getName()+ " got message "+Message);

  	// Need to pick a random index of the message and change to a random char
  	int spot=(int)(Math.random()*Message.length()); // Pick a spot in the String
  	char[] temp=Message.toCharArray();   // Convert to char array
  	char hold=temp[spot];                // Just for reporting purposes
  	temp[spot]=(char)((Math.random()*128)); // Pick random char
  	// State which char was changed to what; removable
  	System.out.println("Spot:"+spot+" Char "+hold+" becomes "+temp[spot]);
  	Message=String.valueOf(temp);     // place in Message
  	// Pass updated message to the monitor
  	try {
      		Control.passMessage(Message);
  	}
  	catch (Exception e) 
	{}
 } // run
} // class

class Monitor {
	private String theMessage;
 	private int nextThread=0,numThreads,numberServed=0;
 	private boolean lock=false;  // used to allow one at a time only

 // Constructor: Get number of threads and initial message
 //    Need # threads for waitUntilDone(), which then allows main
 //    to get back the final version of the message
 public Monitor(String Message,int threads)
 {	numThreads=threads;
  	theMessage=Message;
 }

 // Only one thread may complete at a time
 //   Note synchronized means only one thread may be active here
 public synchronized void waitForTurn(int threadNum)
 {	// wait for next thread in "circle"
  	while (lock || nextThread!=threadNum) {   // wait until lock is false
        	try {
            		wait(1000);  // if not expected thread, wait...
        	}
        	catch (Exception e) {
          		System.out.println("Exception:"+e.toString());
        	}
  	}
  	lock=true;  // Got lock, set to true so no others can pass
 }

 // Holds back main() until all threads allowed through
 public synchronized void waitUntilDone()
 {	while (numberServed<numThreads) {  // wait until all threads pass thru
        	try {
            		wait();  // wait for last notify()
        	}
        	catch (Exception e) {
          		System.err.println("Exception:"+e.toString());
        	}
  	}
 }

 // Get back message from Participant
 //  Must be synchronized because of notify call
 public synchronized void passMessage(String Message)
 {	theMessage=Message;
  	lock=false;
  	numberServed++;
  	nextThread++;
  	notify();        // someone waiting can now come in from waitForTurn()
 }

 // Provide message to Participant
 // Need not be synchronized because
 //      only the process that has the lock can possibly call it
 public String getMessage()
 {	return(theMessage);
 }

}  // Monitor class

// Main class

public class MessagePass {

 static BufferedReader DataIn = new
			    BufferedReader(new InputStreamReader(System.in));

 static int ReadInt() throws java.io.IOException
 {	String Input=DataIn.readLine();
  	return (Integer.valueOf(Input).intValue());
 }

 public static void main(String[] args)
 {	int numberOfThreads=0;
  	// Get number of threads to test
  	try {
   		System.out.print("How many participants?  >");
   		numberOfThreads=ReadInt();
  	}
  	catch (Exception e) {
   		System.out.println("Exception:"+e.toString());
  	}
  	// Get Message
  	String Message="";
  	try {
   		System.out.print("Enter a Message  >");
   		Message=DataIn.readLine();
  	}
  	catch (Exception e) {
   		System.out.println("Exception:"+e.toString());
  	}
  	// Set up the Monitor
  	Monitor Passer=new Monitor(Message,numberOfThreads);
  	// Declare array of threads
  	Participant[] Them=new Participant[numberOfThreads];
  	// Instantiate each thread
  	for (int i=0;i<numberOfThreads;i++)
   		Them[i]=new Participant(String.valueOf(i),Passer);
  	// Start the threads
  	for (int i=0;i<numberOfThreads;i++) {
   		Them[i].start();
  	}
  	// Get Message back
  	Passer.waitUntilDone();
  	Message=Passer.getMessage();
  	try {
   		System.out.println("Final version of message is:"+Message);
   		System.out.println("Hit Return to Exit");
   		System.in.read();
  	}
  	catch (Exception e) 
	{}
 }

} // main class

